from abc import ABC, abstractmethod
from copy import copy
import json
import logging
import os
from ..digital import Digital
from ..array import Array
from ..bits import Bits
from ..clock import wiredefaultclock, wireclock
from coreir import Wireable
from .coreir_utils import (add_non_input_ports, attach_debug_info,
                           check_magma_interface, constant_to_value,
                           get_inst_args, get_module_of_inst,
                           is_clock_or_nested_clock,
                           magma_interface_to_coreir_module_type,
                           magma_port_to_coreir_port, make_cparams,
                           map_genarg, magma_name_to_coreir_select, Slice)
from ..interface import InterfaceKind
from ..is_definition import isdefinition
from ..logging import root_logger
from ..passes import InstanceGraphPass
from ..tuple import Tuple
from .util import get_codegen_debug_info
from magma.clock import ClockTypes, first, get_first_clock, wire_clock_port
from magma.config import get_debug_mode
from magma.protocol_type import MagmaProtocol, MagmaProtocolMeta
from magma.ref import PortViewRef, ArrayRef


# NOTE(rsetaluri): We do not need to set the level of this logger since it has
# already been done in backend/coreir_.py.
_logger = root_logger().getChild("coreir_backend")


def _collect_drivers(value):
    """
    Iterate over value to collect the child drivers, packing slices together
    """
    drivers = []
    start_idx = 0
    for i in range(1, len(value)):
        # If the next value item is not a reference to an array of bits where
        # the array matches the previous item and the index is incremented by
        # one, append the current slice to drivers (may introduce slices of
        # length 1)
        if not (
            isinstance(value[i].name, ArrayRef) and
            issubclass(value[i].name.array.T, Digital) and
            isinstance(value[i - 1].name, ArrayRef) and
            value[i].name.array is value[i - 1].name.array and
            value[i].name.index == value[i - 1].name.index + 1
        ):
            drivers.append(value[start_idx:i])
            start_idx = i
    drivers.append(value[start_idx:])
    return drivers


def _unwrap(x):
    if isinstance(x, MagmaProtocol):
        return x._get_magma_value_()
    if isinstance(x, MagmaProtocolMeta):
        return x._to_magma_()
    return x


class TransformerBase(ABC):
    def __init__(self, backend, opts):
        self.backend = backend
        self.opts = opts
        self.ran = False
        self._children = None

    def run(self):
        if self.ran:
            raise RuntimeError("Can only run transformer once")
        self._children = self.children()
        for child in self._children:
            child.run()
        self.run_self()
        self.ran = True

    @abstractmethod
    def children(self):
        raise NotImplementedError()

    def run_self(self):
        pass


class LeafTransformer(TransformerBase):
    def children(self):
        return []


class DefnOrDeclTransformer(TransformerBase):
    def __init__(self, backend, opts, defn_or_decl):
        super().__init__(backend, opts)
        self.defn_or_decl = defn_or_decl
        self.coreir_module = None

    def children(self):
        if self.defn_or_decl.name in self.backend.modules:
            _logger.debug(f"{self.defn_or_decl} already compiled, skipping")
            self.coreir_module = self.backend.modules[self.defn_or_decl.name]
            return []
        if not isdefinition(self.defn_or_decl):
            return [DeclarationTransformer(self.backend,
                                           self.opts,
                                           self.defn_or_decl)]
        wrapped = getattr(self.defn_or_decl, "wrappedModule", None)
        if wrapped and wrapped.context is self.backend.context:
            return [WrappedTransformer(self.backend,
                                       self.opts,
                                       self.defn_or_decl)]
        return [DefinitionTransformer(self.backend,
                                      self.opts,
                                      self.defn_or_decl)]

    def run_self(self):
        if self.coreir_module:
            return
        self.coreir_module = self._children[0].coreir_module
        self.backend.modules[self.defn_or_decl.name] = self.coreir_module
        if isdefinition(self.defn_or_decl):
            self.defn_or_decl.wrappedModule = self.coreir_module
            libs = copy(self.backend.libs_used)
            self.defn_or_decl.coreir_wrapped_modules_libs_used = libs


class InstanceTransformer(LeafTransformer):
    def __init__(self, backend, opts, inst, defn):
        super().__init__(backend, opts)
        self.inst = inst
        self.defn = defn
        self.coreir_inst_gen = None

    def run_self(self):
        self.coreir_inst_gen = self.run_self_impl()

    def run_self_impl(self):
        _logger.debug(
            f"Compiling instance {(self.inst.name, type(self.inst).name)}"
        )
        defn = type(self.inst)
        lib = self.backend.libs[self.inst.coreir_lib]
        if self.inst.coreir_genargs is None:
            module = get_module_of_inst(self.backend.context, self.inst, lib)
            args = get_inst_args(self.inst)
            args = self.backend.context.new_values(args)
            return lambda m: m.add_module_instance(self.inst.name, module, args)
        generator = lib.generators[defn.coreir_name]
        config_args = {k: v for k, v in self.inst.coreir_configargs.items()}
        config_args = self.backend.context.new_values(config_args)
        gen_args = {k: map_genarg(self.backend.context, v)
                    for k, v in defn.coreir_genargs.items()}
        gen_args = self.backend.context.new_values(gen_args)
        return lambda m: m.add_generator_instance(
            self.inst.name, generator, gen_args, config_args)


class WrappedTransformer(LeafTransformer):
    def __init__(self, backend, opts, defn):
        super().__init__(backend, opts)
        self.defn = defn
        self.coreir_module = self.defn.wrappedModule
        self.backend.libs_used |= self.defn.coreir_wrapped_modules_libs_used


class DefinitionTransformer(TransformerBase):
    def __init__(self, backend, opts, defn):
        super().__init__(backend, opts)
        self.defn = defn
        self.coreir_module = None
        self.decl_tx = DeclarationTransformer(self.backend,
                                              self.opts,
                                              self.defn)
        self.inst_txs = {
            inst: InstanceTransformer(self.backend, self.opts, inst, self.defn)
            for inst in self.defn.instances
        }
        self.clocks = {}
        for clock_type in ClockTypes:
            self.clocks[clock_type] = first(
                get_first_clock(port, clock_type)
                for port in defn.interface.ports.values()
            )

    def children(self):
        pass_ = InstanceGraphPass(self.defn)
        pass_.run()
        deps = [k for k, _ in pass_.tsortedgraph if k is not self.defn]
        children = [DefnOrDeclTransformer(self.backend, self.opts, dep)
                    for dep in deps]
        children += [self.decl_tx]
        children += self.inst_txs.values()
        return children

    def run_self(self):
        _logger.debug(f"Compiling definition {self.defn}")
        self.coreir_module = self.decl_tx.coreir_module
        if self.defn.inline_verilog_strs:
            inline_verilog = "\n\n".join(x[0] for x in
                                         self.defn.inline_verilog_strs)
            connect_references = {}
            for _, inline_value_map in self.defn.inline_verilog_strs:
                for key, value in inline_value_map.items():
                    connect_references[key] = magma_port_to_coreir_port(value)
            self.coreir_module.add_metadata("inline_verilog", json.dumps(
                {"str": inline_verilog,
                 "connect_references": connect_references}
            ))
        for name, module in self.defn.bind_modules.items():
            self.backend.sv_bind_files[name] = module

        self.coreir_module.definition = self.get_coreir_defn()

    def get_coreir_defn(self):
        coreir_defn = self.coreir_module.new_definition()
        coreir_insts = {inst: self.inst_txs[inst].coreir_inst_gen(coreir_defn)
                        for inst in self.defn.instances}
        # If this module was imported from verilog, do not go through the
        # general module construction flow. Instead just attach the verilog
        # source as metadata and return the module.
        if hasattr(self.defn, "verilogFile") and self.defn.verilogFile:
            metadata = json.dumps({"verilog_string": self.defn.verilogFile})
            self.coreir_module.add_metadata("verilog", metadata)
            return coreir_defn
        if self.defn.coreir_lib is not None:
            self.backend.libs_used.add(self.defn.coreir_lib)
        non_input_ports = {}
        for name, port in self.defn.interface.ports.items():
            _logger.debug(f"{name}, {port}, {port.is_output()}")
            add_non_input_ports(non_input_ports, port)
        for inst, coreir_inst in coreir_insts.items():
            if get_codegen_debug_info() and getattr(inst, "debug_info", False):
                attach_debug_info(coreir_inst, inst.debug_info)
            for name, port in inst.interface.ports.items():
                add_non_input_ports(non_input_ports, port)
        for inst in coreir_insts:
            for name, port in inst.interface.ports.items():
                self.connect_non_outputs(coreir_defn, port, non_input_ports)
        for port in self.defn.interface.ports.values():
            self.connect_non_outputs(coreir_defn, port, non_input_ports)
        return coreir_defn

    def connect_non_outputs(self, module_defn, port, non_input_ports):
        # Recurse into non input types that may contain inout children.
        if isinstance(port, Tuple) and not port.is_input() or \
           isinstance(port, Array) and not port.T.is_input():
            for elem in port:
                self.connect_non_outputs(module_defn, elem, non_input_ports)
        elif not port.is_output():
            self.connect(module_defn, port, port.trace(), non_input_ports)

    def get_source(self, port, value, module_defn, non_input_ports):
        port = _unwrap(port)
        value = _unwrap(value)
        if isinstance(value, Wireable):
            return value
        if isinstance(value, Slice):
            return module_defn.select(value.get_coreir_select())
        if isinstance(value, Bits) and value.const():
            return self.const_instance(value, len(value), module_defn)
        if value.anon() and isinstance(value, Array):
            drivers = _collect_drivers(value)
            offset = 0
            for d in drivers:
                d = _unwrap(d)
                if len(d) == 1:
                    # _collect_drivers will introduce a slice of length 1 for
                    # non-slices, so we index them here with 0 to unpack the
                    # extra array dimension
                    self.connect(module_defn, port[offset], d[0],
                                 non_input_ports)
                else:
                    self.connect(module_defn,
                                 Slice(port, offset, offset + len(d)),
                                 Slice(d[0].name.array, d[0].name.index,
                                       d[-1].name.index + 1),
                                 non_input_ports)
                offset += len(d)

            return None
        if isinstance(value, Tuple) and value.anon():
            for p, v in zip(port, value):
                self.connect(module_defn, p, v, non_input_ports)
            return None
        if value.const():
            return self.const_instance(value, None, module_defn)
        if isinstance(value.name, PortViewRef):
            return module_defn.select(
                magma_name_to_coreir_select(value.name))
        return module_defn.select(non_input_ports[value])

    def connect(self, module_defn, port, value, non_input_ports):
        if value is None and is_clock_or_nested_clock(type(port)):
            for type_, default_driver in self.clocks.items():
                if isinstance(port, type_) and default_driver is not None:
                    wire_clock_port(port, type_, default_driver)
                    value = port.value()
                    break
            else:
                # No default clock
                return
        if value is None:
            if port.is_inout():
                return  # skip inouts because they might be conn. as an input.
            if getattr(self.defn, "_ignore_undriven_", False):
                return
            raise Exception(f"Found unconnected port: {port.debug_name}")
        source = self.get_source(port, value, module_defn, non_input_ports)
        if not source:
            return
        sink = module_defn.select(magma_port_to_coreir_port(port))
        module_defn.connect(source, sink)
        if get_codegen_debug_info() and getattr(port, "debug_info", False):
            attach_debug_info(module_defn, port.debug_info, source, sink)

    def const_instance(self, constant, num_bits, module_defn):
        cache_entry = self.backend.constant_cache.setdefault(module_defn, {})
        value = constant_to_value(constant)
        key = (value, num_bits)
        if key in cache_entry:
            return cache_entry[key]
        if num_bits is None:
            config = self.backend.context.new_values({"value": bool(value)})
            name = f"bit_const_{value}_{num_bits}"
            mod = self.backend.libs["corebit"].modules["const"]
            module_defn.add_module_instance(name, mod, config)
        else:
            config = self.backend.context.new_values({"value": value})
            name = f"const_{value}_{num_bits}"
            gen = self.backend.libs["coreir"].generators["const"]
            gen_args = self.backend.context.new_values({"width": num_bits})
            module_defn.add_generator_instance(name, gen, gen_args, config)
        cache_entry[key] = module_defn.select(f"{name}.out")
        return cache_entry[key]


class DeclarationTransformer(LeafTransformer):
    def __init__(self, backend, opts, decl):
        super().__init__(backend, opts)
        self.decl = decl
        self.coreir_module = None

    def run_self(self):
        self.coreir_module = self.run_self_impl()

    def run_self_impl(self):
        self.decl = self.decl
        _logger.debug(f"Compiling declaration {self.decl}")
        if self.decl.coreir_lib is not None:
            self.backend.libs_used.add(self.decl.coreir_lib)
        # These libraries are already available by default in coreir, so we
        # don't need declarations.
        if self.decl.coreir_lib in ["coreir", "corebit", "commonlib"]:
            lib = self.backend.libs[self.decl.coreir_lib]
            if self.decl.coreir_genargs is None:
                return lib.modules[self.decl.coreir_name]
            return lib.generators[self.decl.coreir_name]
        if self.decl.name in self.backend.modules:
            _logger.debug(f"{self.decl} already compiled, skipping")
            return self.backend.modules[self.decl.name]
        if get_debug_mode():
            check_magma_interface(self.decl.interface)
        module_type = magma_interface_to_coreir_module_type(
            self.backend.context, self.decl.interface)
        if isinstance(self.decl.interface, InterfaceKind):
            module_type = self.backend.context.Flip(module_type)
        kwargs = {}
        if hasattr(self.decl, "coreir_config_param_types"):
            param_types = self.decl.coreir_config_param_types
            kwargs["cparams"] = make_cparams(self.backend.context, param_types)
        namespace = self.opts.get("user_namespace",
                                  self.backend.context.global_namespace)
        coreir_module = namespace.new_module(
            self.decl.coreir_name, module_type, **kwargs)
        if get_codegen_debug_info() and self.decl.debug_info:
            attach_debug_info(coreir_module, self.decl.debug_info)
        return coreir_module



import hwtypes


_SPECIFIERS = [("h", 16), ("o", 8), ("b", 2), ("d", 10)]


def int_const_str_to_int(value):
    """Parses a verilog integer constant (as a string) and returns an int"""
    # Default no specifier.
    size = 32
    base = 10
    signed = False
    if "'" in value:
        # Parse literal specifier.
        size, value = value.split("'")
        size = 32
        if size != "":
            size = int(size)
        signed = False
        if "s" in value:
            signed = True
            value = value.replace("s", "")
        for specifier, base in _SPECIFIERS:
            if specifier in value:  # remove specifier
                value = value.replace(specifier, "")
                break
        else:
            base = 10  # default base 10
    value = int(value, base)
    if signed:
        return hwtypes.SIntVector[size](value).as_uint()
    return hwtypes.UIntVector[size](value).as_uint()



"""
Infrastructure for writing magma tests
"""
import filecmp
import os
import sys
import difflib
from ..config import get_debug_mode, set_debug_mode


def check_files_equal(callee_file, file1_name, file2_name):
    """
    Check if file1 == file2 where file1 and file2 are in the same path as
    callee_file (typically __file__ for the script calling this function)
    """
    file_path = os.path.dirname(callee_file)
    file1_path = os.path.join(file_path, file1_name)
    file2_path = os.path.join(file_path, file2_name)
    print(file1_path, file2_path)
    result = filecmp.cmp(file1_path, file2_path, shallow=False)
    if not result:
        with open(file1_path, "r") as file1:
            with open(file2_path, "r") as file2:
                diff = difflib.unified_diff(
                    file2.readlines(),
                    file1.readlines(),
                    fromfile=file2_name,
                    tofile=file1_name,
                )
                for line in diff:
                    sys.stderr.write(line)
    return result


class _MagmaDebugSection:
    def __init__(self):
        self.__restore = get_debug_mode()

    def __enter__(self):
        set_debug_mode(True)

    def __exit__(self, typ, value, traceback):
        set_debug_mode(self.__restore)


def magma_debug_section():
    return _MagmaDebugSection()


def has_log(caplog, level=None, msg=None):
    if level is None and msg is None:
        return len(caplog.records) > 0
    if level and not msg:
        gen = (log.levelname == level for log in caplog.records)
    elif msg and not level:
        gen = (log.msg == msg for log in caplog.records)
    else:
        gen = (log.levelname == level and log.msg == msg
               for log in caplog.records)
    return any(gen)


def has_error(caplog, msg=None):
    return has_log(caplog, "ERROR", msg)


def has_warning(caplog, msg=None):
    return has_log(caplog, "WARNING", msg)



import inspect
import typing
import ast
from ..ast_utils import get_ast
from ..protocol_type import MagmaProtocol
import astor
import magma.ast_utils as ast_utils
from magma.debug import debug_info
import functools
import magma as m
from magma.ssa import convert_tree_to_ssa
from magma.config import get_debug_mode
from magma.t import Type
from collections import Counter
import itertools

from ast_tools.stack import _SKIP_FRAME_DEBUG_STMT, SymbolTable
from ast_tools import gen_free_name


def _get_magma_value(val):
    if isinstance(val, MagmaProtocol):
        return val._get_magma_value_()
    return val

class RewriteSelfAttributes(ast.NodeTransformer):
    def __init__(self, initial_value_map):
        self.initial_value_map = initial_value_map
        self.calls_seen = []

    def visit_Attribute(self, node):
        if isinstance(node.value, ast.Name) and node.value.id == "self":
            return ast.Name(f"self_{node.attr}_I", node.ctx)
        return self.generic_visit(node)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute) and \
                isinstance(node.func.value, ast.Name) and \
                node.func.value.id == "self":
            attr = node.func.attr
            assert attr in self.initial_value_map, \
                "Reference to self that was not initialized"
            ret = []
            ports = self.initial_value_map[attr][3].interface.inputs()
            for name, value in zip(ports, node.args):
                name = _get_magma_value(name)
                ret.append(ast.parse(
                    f"self_{attr}_{name} = {astor.to_source(value).rstrip()}"
                ).body[0])
            self.calls_seen.extend(ret)
            func = astor.to_source(node.func).rstrip()
            outputs = self.initial_value_map[attr][3].interface.outputs()
            if len(outputs) == 1:
                name = outputs[0]
                name = _get_magma_value(name)
                return ast.Name(f"self_{attr}_{name}", ast.Load())
            else:
                assert outputs, "Expected module with at least one output"
                return ast.Tuple([ast.Name(f"self_{attr}_{_get_magma_value(output)}",
                                           ast.Load()) for output in outputs],
                                 ast.Load())
        elif (isinstance(node.func, ast.Attribute) and
                isinstance(node.func.value, ast.Attribute) and
                isinstance(node.func.value.value, ast.Name) and
                node.func.value.value.id == "self" and
                node.func.attr == "prev"):
            return ast.Name(f"self_{node.func.value.attr}_O", ast.Load())
        else:
            self.generic_visit(node)
        return node

    def visit_If(self, node):
        node.test = self.visit(node.test)
        for attr in ["body", "orelse"]:
            new_body = []
            for item in getattr(node, attr):
                new_body.append(self.visit(item))
            if self.calls_seen:
                new_body = self.calls_seen + new_body
                self.calls_seen = []
            setattr(node, attr, new_body)
        return node


    # def visit(self, node):
    #     node = super().visit(node)
    #     if hasattr(node, 'body'):
    #         new_body = self.calls_seen
    #         self.calls_seen = []
    #         for item in node.body:
    #             if isinstance(item, list):
    #                 new_body.extend(item)
    #             else:
    #                 new_body.append(item)
    #         node.body = new_body
    #     return node


class RewriteReturn(ast.NodeTransformer):
    def __init__(self, initial_value_map):
        self.initial_value_map = initial_value_map

    def visit_Return(self, node):
        elts = []
        for name, (value, type_, eval_type, eval_value) in self.initial_value_map.items():
            if isinstance(eval_type, m.Kind):
                elts.append(ast.Name(f"self_{name}_I", ast.Load()))
            else:
                for port in eval_value.interface.inputs():
                    if isinstance(port, (m.Clock, m.AsyncReset)):
                        continue
                    port = _get_magma_value(port)
                    elts.append(ast.Name(f"self_{name}_{port}", ast.Load()))
        if isinstance(node.value, ast.Tuple):
            elts.extend(node.value.elts)
        else:
            elts.append(node.value)
        if len(elts) == 1:
            node.value == 0
        else:
            node.value = ast.Tuple(elts, ast.Load())
        return node


def get_initial_value_map(init_func, defn_env):
    """
    Parses __init__ funciton of the form

        def __init__(self):
            self.x = m.bits(0, 2)
            self.y = m.bits(0, 4)

    Returns a map from instance attributes to their initial values
    For the above example,

        {'x': <_ast.Call object at 0x109c17fd0>,
         'y': <_ast.Call object at 0x109c17470>}

    TODO: Should we allow arbitrary code in the definition? For now, the user
    can write any meta code outside the class definition, which makes the
    design of this simpler
    """
    initial_value_map = {}
    if init_func is object.__init__:
        # Handle case when no __init__
        return initial_value_map
    init_def = get_ast(init_func).body[0]
    # init_def = ExecuteEscapedPythonExpressions(defn_env).visit(init_def)
    init_def = SpecializeConstantInts(defn_env).visit(init_def)
    for stmt in init_def.body:
        # We only support basic assignments of the form
        #     self.x: m.Bits(2) = m.bits(0, 2)
        assert isinstance(stmt, ast.AnnAssign)
        assert isinstance(stmt.target, ast.Attribute)
        assert isinstance(stmt.target.value, ast.Name)
        assert stmt.target.value.id == "self"
        # TODO: Should we deal with multiple assignments? For now we take the
        # last one
        class Tmp(m.Circuit):
            eval_type = eval(astor.to_source(stmt.annotation).rstrip(), defn_env)
            eval_value = eval(astor.to_source(stmt.value).rstrip(), defn_env)
        initial_value_map[stmt.target.attr] = (stmt.value, stmt.annotation,
                                               Tmp.eval_type,
                                               Tmp.eval_value)
    return initial_value_map


circuit_definition_template = """
from magma import Bit, Array, Tuple, Product, Bits, SInt, UInt, IO

def make_{circuit_name}(combinational):
    class {circuit_name}({magma_name}.Circuit):
        io = IO({io_args})

        {register_instances}
        @combinational
        def {circuit_name}_comb({circuit_combinational_args}) -> ({circuit_combinational_output_type}):
            {circuit_combinational_body}
        comb_out = {circuit_name}_comb({circuit_combinational_call_args})
        {comb_out_wiring}
    return {circuit_name}
"""


def gen_array_str(eval_type, eval_value, reset_type, magma_name):
    arr = [gen_reg_inst_str(eval_type.T, eval_value[i], reset_type, magma_name) for i
           in range(len(eval_type))]
    return f"{magma_name}.join([{', '.join(arr)}])"


def gen_product_str(eval_type, eval_value, reset_type, magma_name):
    t = [
        f"'{key}':" + gen_reg_inst_str(eval_type[key], eval_value[key], reset_type, magma_name)
        for key in eval_type.keys()
    ]
    return f"{{{', '.join(t)}}}"


def gen_reg_inst_str(eval_type, eval_value, reset_type, magma_name):
    if issubclass(eval_type, m.Digital):
        n = None
        if isinstance(eval_value, (bool, int)):
            init = bool(eval_value)
        else:
            assert eval_value.name.name in ["GND", "VCC"], eval_value.name
            init = 0 if eval_value.name.name == "GND" else 1
    elif issubclass(eval_type, m.Array) and \
            issubclass(eval_type.T, m.Digital):
        n = len(eval_type)
        init = int(eval_value)
    elif issubclass(eval_type, m.Array):
        return f"{gen_array_str(eval_type, eval_value, reset_type, magma_name)}"
    elif issubclass(eval_type, m.Product):
        return f"{gen_product_str(eval_type, eval_value, reset_type, magma_name)}"
    else:
        raise NotImplementedError((eval_type))
    reset_arg = {
        None: "",
        m.AsyncReset: "has_async_reset=True",
        m.AsyncResetN: "has_async_resetn=True",
        m.Reset: "has_reset=True",
        m.ResetN: "has_resetn=True"
    }[reset_type]
    return f"DefineRegister({n}, init={init}, {reset_arg})()"


def gen_register_instances(initial_value_map, reset_type, magma_name):
    """
    Generates a sequence of statements to instance a set of registers from
    `initial_value_map`.

    For example,

        initial_value_map = {'x': <_ast.Call object at 0x109c17fd0>,
                             'y': <_ast.Call object at 0x109c17470>}

    will generate

        ['x = Register(2, init=0)', 'y = Register(4, init=0)']

    """
    register_instances = []
    for name, (value, type_, eval_type, eval_value) in initial_value_map.items():
        if isinstance(eval_type, m.Kind):
            reg_inst_str = gen_reg_inst_str(eval_type, eval_value,
                                            reset_type, magma_name)
            register_instances.append(f"{name} = {reg_inst_str}")
        else:
            value = astor.to_source(value).rstrip()
            register_instances.append(f"{name} = {value}")
    return register_instances


class TypeNamer:
    def __init__(self, tree, env):
        self.tree = tree
        self.env = env
        self.cache = {}

    def __call__(self, t):
        try:
            return self.cache[t]
        except KeyError:
            pass

        # Try to keep the name the same
        for k, v in self.env.items():
            if v is t:
                return self.cache.setdefault(t, k)

        # Maybe we can exec the repr?
        try:
            s = eval(repr(t), env)
        except:
            pass
        else:
            if s is t:
                return self.cache.setdefault(t, repr(t))

        # Or the str?
        try:
            s = eval(str(t), env)
        except:
            pass
        else:
            if s is t:
                return self.cache.setdefault(t, str(t))

        # give up generate a fesh one
        n = self.cache.setdefault(t, gen_free_name(self.tree, self.env, 'T'))
        self.env[n] = t
        return n


def get_io_from_annotations(f, tree, defn_env):
    inputs = []
    outputs = []
    tuple_out = False
    namer = TypeNamer(tree, defn_env)
    for k, v in f.__annotations__.items():
        if k == 'return':
            if isinstance(v, tuple):
                tuple_out = True
                for idx, t in enumerate(v):
                    t_name = namer(t)
                    outputs.append((f'O{idx}', t_name))
            else:
                t_name = namer(v)
                outputs.append(('O', t_name))
        else:
            t_name = namer(v)
            inputs.append((k, t_name))

    return inputs, outputs, tuple_out


def gen_io_args(inputs, outputs, reset_type, magma_name):
    io_args = []
    for name, type_ in inputs:
        io_args.append(f"{name}={magma_name}.In({type_})")
    io_args.append(f"CLK={magma_name}.In({magma_name}.Clock)")
    if reset_type is not None:
        port_name, port_type = {
            m.AsyncReset: ("ASYNCRESET", "AsyncReset"),
            m.AsyncResetN: ("ASYNCRESETN", "AsyncResetN"),
            m.Reset: ("RESET", "Reset"), 
            m.ResetN: ("RESETN", "ResetN")
        }[reset_type]
        io_args.append(f"{port_name}={magma_name}.In({magma_name}.{port_type})")
    for name, type_ in outputs:
        io_args.append(f"{name}={magma_name}.Out({type_})")
    return ', '.join(io_args)


class EscapedExpression(ast.AST):
    def __init__(self, value, orig):
        self.value = value
        self.orig = orig


class ExecuteEscapedPythonExpressions(ast.NodeTransformer):
    def __init__(self, defn_env):
        self.defn_env = defn_env

    def visit_List(self, node):
        assert len(node.elts) == 1, "Expected only list literatals that " \
            "contain a single expression"
        # FIXME: Hack to prevent instancing logic from polluting currently open
        # definition
        m.DefineCircuit("tmp")
        result = eval(astor.to_source(node.elts[0]).rstrip(), self.defn_env)
        m.EndCircuit()
        return EscapedExpression(result, node)


class SpecializeConstantInts(ast.NodeTransformer):
    def __init__(self, defn_env):
        self.defn_env = defn_env

    def visit_Name(self, node):
        if node.id in self.defn_env:
            value = self.defn_env[node.id]
            if isinstance(value, int):
                return ast.Num(value)
        return node


def gen_product_call_args(name, comb_out_count, eval_type, keys, comb_out_wiring, magma_name):
    call_args = []
    for key in eval_type.keys():
        if isinstance(eval_type[key], m.ProductKind):
            call_args.append(
                f"{key}="
                + gen_product_call_args(
                    name, comb_out_count, eval_type[key], keys + [key], comb_out_wiring, magma_name
                )
            )
        else:
            arrays = "".join([f"['{k}']" for k in keys + [key]])
            dots = ".".join(keys + [key])
            comb_out_wiring.append(
                f"{name}{arrays}.I @= comb_out[{comb_out_count}].{dots} \n"
            )
            call_args.append(f"{key}={name}{arrays}.O")
    return f"{magma_name}.namedtuple(" + ", ".join(call_args) + ")"


def _sequential(
        defn_env: dict,
        reset_type: Type,
        cls,
        combinational_decorator: typing.Callable):
    # if not inspect.isclass(cls):
    #     raise ValueError("sequential decorator only works with classes")


    initial_value_map = get_initial_value_map(cls.__init__, defn_env)

    call_def = get_ast(cls.__call__).body[0]
    magma_name = gen_free_name(call_def, defn_env, 'm')
    defn_env[magma_name] = m

    inputs, outputs, tuple_out = get_io_from_annotations(cls.__call__, call_def, defn_env)
    io_args = gen_io_args(inputs, outputs, reset_type, magma_name)

    circuit_combinational_output_type = []
    circuit_combinational_args = []
    circuit_combinational_call_args = []
    comb_out_wiring = []
    for name, type_ in inputs:
        circuit_combinational_args.append(f"{name}: {type_}")
        circuit_combinational_call_args.append(f"{name}=io.{name}")

    comb_out_count = 0
    for name, (value, type_, eval_type, eval_value) in initial_value_map.items():
        if isinstance(eval_type, m.Kind):
            type_ = astor.to_source(type_).rstrip()
            circuit_combinational_args.append(f"self_{name}_O: {type_}")
            circuit_combinational_output_type.append(f"{type_}")
            if isinstance(eval_type, m.ProductKind):
                call_args = gen_product_call_args(
                    name, comb_out_count, eval_type, [], comb_out_wiring, magma_name
                )
                circuit_combinational_call_args.append(f"self_{name}_O=" + call_args)
            else:
                circuit_combinational_call_args.append(f"self_{name}_O={name}")
                comb_out_wiring.append(f"{name}.I @= comb_out[{comb_out_count}]\n")
            comb_out_count += 1
        else:
            for key, value in eval_value.interface.ports.items():
                if isinstance(value, (m.Clock, m.AsyncReset)):
                    continue
                t_name = gen_free_name(call_def, defn_env, 'T')
                defn_env[t_name] = type(value)
                if value.is_output():
                    circuit_combinational_args.append(f"self_{name}_{key}: {t_name}")
                    circuit_combinational_call_args.append(f"self_{name}_{key}={name}.{key}")
                if value.is_input():
                    circuit_combinational_output_type.append(t_name)
                    comb_out_wiring.append(f"{name}.{key} @= comb_out[{comb_out_count}]\n")
                    comb_out_count += 1

    circuit_combinational_args = ', '.join(circuit_combinational_args)
    circuit_combinational_call_args = ', '.join(circuit_combinational_call_args)

    if tuple_out:
        for i, (n, t) in enumerate(outputs):
            circuit_combinational_output_type.append(t)
            comb_out_wiring.append(f"io.{n} @= comb_out[{comb_out_count + i}]\n")
    else:
        assert len(outputs) == 1
        n, t = outputs[0]
        circuit_combinational_output_type.append(t)
        # Handle case when no registers, so only one output
        index_str = ""
        if comb_out_count > 0:
            index_str = f"[{comb_out_count}]"
        comb_out_wiring.append(f"io.O @= comb_out{index_str}\n")

    tab = 4 * ' '
    comb_out_wiring = (2 * tab).join(comb_out_wiring)
    circuit_combinational_output_type = ', '.join(circuit_combinational_output_type)
    circuit_combinational_body = []
    for name, (value, type_, eval_type, eval_value) in initial_value_map.items():
        if isinstance(eval_type, m.Kind):
            circuit_combinational_body.append(
                f"self_{name}_I = self_{name}_O")
    for stmt in call_def.body:
        rewriter = RewriteSelfAttributes(initial_value_map)
        stmt = rewriter.visit(stmt)
        code = [stmt]
        if rewriter.calls_seen:
            code = rewriter.calls_seen + code
        stmt = RewriteReturn(initial_value_map).visit(stmt)
        for stmt in code:
            for line in astor.to_source(stmt).rstrip().splitlines():
                circuit_combinational_body.append(line)

    circuit_combinational_body = ('\n' + 3*tab).join(circuit_combinational_body)
    register_instances = gen_register_instances(initial_value_map, reset_type, magma_name)
    register_instances = ('\n' + 2*tab).join(register_instances)

    circuit_definition_str = circuit_definition_template.format(
        circuit_name=cls.__name__,
        io_args=io_args,
        register_instances=register_instances,
        circuit_combinational_args=circuit_combinational_args,
        circuit_combinational_output_type=circuit_combinational_output_type,
        circuit_combinational_body=circuit_combinational_body,
        circuit_combinational_call_args=circuit_combinational_call_args,
        comb_out_wiring=comb_out_wiring,
        magma_name=magma_name,
    )
    tree = ast.parse(circuit_definition_str)
    if "DefineRegister" not in defn_env:
        tree = ast.Module([
            ast.parse("from mantle import DefineRegister").body[0],
        ] + tree.body)

    circuit_def_constructor = ast_utils.compile_function_to_file(tree, 'make_' + cls.__name__, defn_env)
    circuit_def = circuit_def_constructor(combinational_decorator)

    if get_debug_mode() and getattr(circuit_def, "debug_info", False):
        circuit_def.debug_info = debug_info(circuit_def.debug_info.filename,
                                            circuit_def.debug_info.lineno,
                                            inspect.getmodule(cls))

    return circuit_def

def sequential(
        cls=None,
        async_reset=None,
        reset_type=None,
        *,
        decorators: typing.Optional[typing.Sequence[typing.Callable]] = None,
        env: SymbolTable = None):

    exec(_SKIP_FRAME_DEBUG_STMT)
    if (async_reset is not None or decorators is not None or 
            reset_type is not None):
        if async_reset is not None and not isinstance(async_reset, bool):
            raise TypeError("async_reset param should be a bool")
        if async_reset:
            # will override reset_type if it is also passed
            reset_type = m.AsyncReset
        if cls is not None:
            raise ValueError("Should not explicitly pass first argument (cls) "
                             f"to sequential, use kwargs instead")
        if decorators is None:
            decorators = ()

        def wrapped(cls):
            exec(_SKIP_FRAME_DEBUG_STMT)
            nonlocal decorators
            decorators = list(itertools.chain(decorators, [wrapped]))
            wrapped_sequential = ast_utils.inspect_enclosing_env(
                    _sequential,
                    decorators=decorators,
                    st=env
            )
            combinational = m.circuit.combinational(decorators=decorators, env=env)
            return wrapped_sequential(reset_type, cls, combinational)
        return wrapped
    else:
        assert cls is not None
        wrapped_sequential = ast_utils.inspect_enclosing_env(
                _sequential,
                decorators=[sequential],
                st=env
        )
        combinational = m.circuit.combinational(decorators=[sequential], env=env)
        return wrapped_sequential(m.AsyncReset, cls, combinational)
